#include <BLEDevice.h>
#include <BLEServer.h>
#include <BLEUtils.h>
#include <BLE2902.h>
#include <Wire.h>
#include <Adafruit_Sensor.h>
#include <Adafruit_BME280.h>


Adafruit_BME280 bme;      // I2C

float tempC;
float pressure;
float humidity;
bool deviceConnected = false;
bool oldDeviceConnected = false;
char msg[10];
unsigned long lastTime = 0;
unsigned long nextTime = 5000;

// Generated by https://www.uuidgenerator.net/
#define SERVICE_UUID  "ecc34a9c-e4f4-11ec-8fea-0242ac120002"
#define TEMPCC_UUID   "ecc34ca4-e4f4-11ec-8fea-0242ac120002"
#define PRESSURE_UUID "ecc34dda-e4f4-11ec-8fea-0242ac120002"
#define HUMIDITY_UUID "ecc34ef2-e4f4-11ec-8fea-0242ac120002"

BLECharacteristic* pTempCharacteristic;
BLECharacteristic* pPressCharacteristic;
BLECharacteristic* pHumidityCharacteristic;

//Setup callbacks onConnect and onDisconnect
class MyServerCallbacks: public BLEServerCallbacks {
    void onConnect(BLEServer* pServer) {
      deviceConnected = true;
    };
    void onDisconnect(BLEServer* pServer) {
      deviceConnected = false;
    }
};

void initBME() {
  if (!bme.begin(0x76)) {
    Serial.println("Could not find a valid BME280 sensor, check wiring!");
    while (1);
  }
}

void setup() {
  // Start serial communication
  Serial.begin(115200);

  // Init BME Sensor
  if (!bme.begin(0x76)) {
    Serial.println("Could not find a valid BME280 sensor, check wiring!");
    while (1);
  }

  BLEDevice::init("ESP32_BME280");                 // Create the BLE Device
  BLEServer *pServer = BLEDevice::createServer();  // Create the BLE Server
  pServer->setCallbacks(new MyServerCallbacks());
  BLEService *bmeService = pServer->createService(SERVICE_UUID); // Create the BLE Service

  // Create BLE Characteristics and Create a BLE Descriptor
  //--- Temperature
  pTempCharacteristic = bmeService->createCharacteristic(
                          TEMPCC_UUID,
                          BLECharacteristic::PROPERTY_READ   |
                          BLECharacteristic::PROPERTY_NOTIFY
                        );
  pTempCharacteristic->addDescriptor(new BLE2902());

  //--- Pressure
  pPressCharacteristic = bmeService->createCharacteristic(
                           PRESSURE_UUID,
                           BLECharacteristic::PROPERTY_READ   |
                           BLECharacteristic::PROPERTY_NOTIFY
                         );
  pPressCharacteristic->addDescriptor(new BLE2902());

  //--- Humidity
  pHumidityCharacteristic = bmeService->createCharacteristic(
                              HUMIDITY_UUID,
                              BLECharacteristic::PROPERTY_READ   |
                              BLECharacteristic::PROPERTY_NOTIFY
                            );
  pHumidityCharacteristic->addDescriptor(new BLE2902());

  bmeService->start();   // Start the service

  //--- Start advertising ---
  BLEAdvertising *pAdvertising = BLEDevice::getAdvertising();
  pAdvertising->addServiceUUID(SERVICE_UUID);
  pAdvertising->setScanResponse(true);
  pAdvertising->setMinPreferred(0x06);  // functions that help with iPhone connections issue
  pAdvertising->setMaxPreferred(0x12);
  BLEDevice::startAdvertising();
  Serial.println("Waiting a client connection to notify...");
}

void loop() {
  if (deviceConnected) {
    if ((millis() - lastTime) > nextTime) {
      tempC = bme.readTemperature();
      pressure = bme.readPressure() / 100.0;
      humidity = bme.readHumidity();
      dtostrf(tempC, -6, 1, msg);
      //Set temperature Characteristic value and notify connected client
      pTempCharacteristic->setValue(msg);
      pTempCharacteristic->notify();
      Serial.print("Temperature: ");
      Serial.print(tempC);
      Serial.print(" ºC");
      dtostrf(pressure, -4, 0, msg);
      //Set pressure Characteristic value and notify connected client
      pPressCharacteristic->setValue(msg);
      pPressCharacteristic->notify();
      Serial.print(" - Pressure: ");
      Serial.print(pressure);
      Serial.print(" hPa");
      dtostrf(humidity,-4, 1, msg);
      pHumidityCharacteristic->setValue(msg);
      pHumidityCharacteristic->notify();
      Serial.print(" - Humidity: ");
      Serial.print(humidity);
      Serial.println(" %");
      lastTime += nextTime;
    }
  }
  // disconnecting
  if (!deviceConnected && oldDeviceConnected) {
    delay(1000); // give the bluetooth stack the chance to get things ready
    BLEDevice::startAdvertising(); // restart advertising
    Serial.println("start advertising");
    oldDeviceConnected = deviceConnected;
  }
  // connecting
  if (deviceConnected && !oldDeviceConnected) {
    Serial.println("Client connected");
    oldDeviceConnected = deviceConnected;
  }
}
